import os
import re
from typing import Optional, NamedTuple, cast, Callable, Generator, Union

import pytest
from _pytest.fixtures import FixtureRequest
from _pytest.python import Function
from _pytest.runner import CallInfo

from ._types import (
    ImageFileType,
    ImageRegressionCallableType,
    ImageDiffCallableType,
    PathType,
    ImageDiffColorModeType,
    ColorType,
)
from .helpers import get_test_info, build_filename, image_save, ensure_dirs, temp_file
from .image_diff import _diff, ALPHA_COLOR

try:
    from .splinter import screenshot_regression  # noqa
except ImportError:
    pass


@pytest.hookimpl(tryfirst=True, hookwrapper=True)  # type: ignore
def pytest_runtest_makereport(item: Function, call: CallInfo):  # type: ignore
    # execute all other hooks to obtain the report object
    outcome = yield
    rep = outcome.get_result()

    # set a report attribute for each phase of a call, which can
    # be "setup", "call", "teardown"

    setattr(item, "rep_" + rep.when, rep)


@pytest.fixture(scope="session")
def image_diff_threshold() -> float:
    """
    Set default threshold differences of images. By default - 0.001
    """
    return 0.001


@pytest.fixture(scope="session")
def image_diff_throw_exception() -> bool:
    """
    Set default throw exception. By default - True
    """
    return True


@pytest.fixture(scope="session")
def image_diff_root(request: FixtureRequest) -> PathType:
    """
    Root path for storing diff images. By default - `request.config.rootdir`
    """
    return str(request.config.rootdir)  # type: ignore


@pytest.fixture(scope="session")  # pragma: no cover
def image_diff_dir(image_diff_root: str) -> PathType:
    """
    Path for store diff images. by default - '{image_diff_root}.tests/image_diff/'
    """
    return os.path.join(image_diff_root, ".tests", "image_diff")


@pytest.fixture(scope="session")
def image_diff_reference_dir(image_diff_root: str) -> PathType:
    """
    Path for store reference images
    """
    return os.path.join(image_diff_root, "image_diff_reference")


@pytest.fixture(scope="session")
def image_diff_color_mode() -> ImageDiffColorModeType:
    """
    Normalize color mode. RGB, RGBA, None. Default None - do nothing
    """
    return None


@pytest.fixture(scope="session")
def image_diff_alpha_color() -> ColorType:
    """
    For alpha channel use fake color
    """
    return ALPHA_COLOR


class DiffInfo(NamedTuple):
    diff_name: str
    image_name: str
    reference_name: str


DiffInfoCallableType = Callable[[ImageFileType, Optional[str]], DiffInfo]


class DiffCompareResult:
    value: float
    thresh: float
    diff_path: Optional[str]

    CompareType = Union[int, float, "DiffCompareResult"]

    def __init__(self, value: float, thresh: float, diff_path: Optional[str] = None):
        self.value = value
        self.thresh = thresh
        self.diff_path = diff_path

    def __repr__(self) -> str:
        return f"diff: {round(self.value, 4)}; thresh: {self.thresh}; diff path: {self.diff_path}"

    def __bool__(self) -> bool:
        return self.value <= self.thresh

    def __gt__(self, other: CompareType) -> bool:
        return self.value > other

    def __ge__(self, other: CompareType) -> bool:
        return self.value >= other

    def __lt__(self, other: CompareType) -> bool:
        return self.value < other

    def __le__(self, other: CompareType) -> bool:
        return self.value <= other


@pytest.fixture(scope="function")
def _image_diff_info(
    request: FixtureRequest, image_diff_reference_dir: str, image_diff_dir: str
) -> DiffInfoCallableType:
    """
    For internal use
    """

    def _factory(image: ImageFileType, suffix: Optional[str] = None) -> DiffInfo:
        test_info = get_test_info(request)
        class_name = re.sub(r"[^\w_. -]", "_", test_info.class_name)
        test_name = re.sub(r"[^\w_. -]", "_", test_info.test_name)
        if suffix is None:
            # Todo enumerate every call
            suffix = ""

        reference_dir = os.path.join(image_diff_reference_dir, class_name)
        screenshot_dir = os.path.join(image_diff_dir, class_name)
        reference_name = os.path.join(
            reference_dir, build_filename(test_name, suffix=suffix + "-reference.png")
        )
        image_name = os.path.join(
            screenshot_dir, build_filename(test_name, suffix=suffix + "-screenshot.png")
        )
        diff_name = os.path.join(
            screenshot_dir, build_filename(test_name, suffix=suffix + "-diff.png")
        )
        return DiffInfo(
            diff_name=diff_name, image_name=image_name, reference_name=reference_name
        )

    return _factory


@pytest.fixture(scope="function")
def image_regression(
    request: FixtureRequest,
    _image_diff_info: DiffInfoCallableType,
    image_diff_threshold: float,
    image_diff_throw_exception: bool,
    image_diff_color_mode: ImageDiffColorModeType,
    image_diff_alpha_color: ColorType,
) -> Generator[ImageRegressionCallableType, None, None]:
    """
    Check regression image.

    :param image: `PIL.Image` or `PathLike` or `io.BinaryIO`
    :param threshold: float, by default from `image_diff_threshold`
    :param suffix: str, need for multiple checks  by one test
    :return: bool
    """

    def _factory(
        image: ImageFileType,
        threshold: Optional[float] = None,
        suffix: Optional[str] = None,
        throw_exception: Optional[bool] = None,
        color_mode: Optional[ImageDiffColorModeType] = None,
        alpha_color: Optional[ColorType] = None,
    ) -> "DiffCompareResult":
        if threshold is None:
            threshold = image_diff_threshold
        if throw_exception is None:
            throw_exception = image_diff_throw_exception

        if color_mode is None:
            color_mode = image_diff_color_mode
        if alpha_color is None:
            alpha_color = image_diff_alpha_color

        diff_info = _image_diff_info(image, suffix)
        reference_name = diff_info.reference_name
        diff_name = diff_info.diff_name
        image_name = diff_info.image_name
        if not os.path.exists(reference_name):
            ensure_dirs(reference_name)
            image_save(image, reference_name)
            return DiffCompareResult(0, threshold)

        def _cleanup() -> None:
            if request.node.rep_setup.passed and request.node.rep_call.failed:
                # Do not cleanup regression artifacts
                return
            for f in [image_name, diff_name]:
                if os.path.exists(f):
                    try:
                        os.unlink(f)
                    except PermissionError:
                        # Opps, have windows issue
                        pass

        request.addfinalizer(_cleanup)

        ensure_dirs(diff_name)
        ensure_dirs(image_name)
        image_save(image, image_name)
        diff_ratio = _diff(
            reference_name,
            image_name,
            diff_name,
            color_mode=color_mode,
            alpha_color=alpha_color,
        )
        cond = DiffCompareResult(diff_ratio, threshold, diff_name)
        if throw_exception:
            assert cond, "Image not equals!"  # noqa
        return cond

    yield _factory


@pytest.fixture(scope="function")
def image_diff(
    request: FixtureRequest,
    _image_diff_info: DiffInfoCallableType,
    image_diff_threshold: float,
    image_diff_throw_exception: bool,
    image_diff_color_mode: ImageDiffColorModeType,
    image_diff_alpha_color: ColorType,
) -> Generator[ImageDiffCallableType, None, None]:
    """
    Compare two image

    :param image: `PIL.Image` or `PathLike` or `io.BinaryIO`
    :param image2: `PIL.Image` or `PathLike` or `io.BinaryIO`
    :param threshold: float, by default from `image_diff_threshold`
    :param suffix: str, need for multiple checks  by one test
    :return: bool
    """

    def _factory(
        image: ImageFileType,
        image_2: ImageFileType,
        threshold: Optional[float] = None,
        suffix: Optional[str] = None,
        throw_exception: Optional[bool] = None,
        color_mode: Optional[ImageDiffColorModeType] = None,
        alpha_color: Optional[ColorType] = None,
    ) -> DiffCompareResult:
        if threshold is None:
            threshold = image_diff_threshold
        if throw_exception is None:
            throw_exception = image_diff_throw_exception

        if color_mode is None:
            color_mode = image_diff_color_mode
        if alpha_color is None:
            alpha_color = image_diff_alpha_color

        _info = _image_diff_info(image, suffix)
        diff_path = cast(str, _info.diff_name)

        ensure_dirs(diff_path)

        def _cleanup() -> None:
            if request.node.rep_setup.passed and request.node.rep_call.failed:
                # Do not cleanup regression artifacts
                return
            for f in [diff_path]:
                if os.path.exists(f):
                    os.unlink(f)

        request.addfinalizer(_cleanup)

        with temp_file(suffix=".png") as image_temp_file, temp_file(
            suffix=".png"
        ) as image_2_temp_file:
            image_save(image, path=image_temp_file)
            image_save(image_2, path=image_2_temp_file)
            diff_ratio = _diff(
                image_temp_file,
                image_2_temp_file,
                diff_path,
                color_mode=color_mode,
                alpha_color=alpha_color,
            )
        cond = DiffCompareResult(diff_ratio, threshold, diff_path)
        if throw_exception:
            assert cond, "Image not equals! See %s" % diff_path  # noqa
        return cond

    yield _factory



import os
import pathlib
import shutil
import typing
from contextlib import contextmanager
from tempfile import NamedTemporaryFile

from PIL.Image import Image
from _pytest import junitxml
from _pytest.fixtures import FixtureRequest

from pytest_image_diff._types import ImageFileType, PathType


def build_filename(
    name: str, suffix: str = "", prefix: str = "", max_length: int = 128
) -> str:
    """
    >>> build_filename("name")
    'name'
    >>> build_filename("name", "suffix", "prefix")
    'prefix-name-suffix'
    >>> build_filename("loooooooong_nameeeeee", "suffix", max_length=20)
    'loooooooo-suffix'
    """
    return "-".join(
        filter(
            None, [prefix, name[: max_length - len(suffix) - len(prefix) - 5], suffix]
        )
    )


@contextmanager
def temp_file(suffix: typing.Optional[str] = None) -> typing.Iterator[pathlib.Path]:
    with NamedTemporaryFile(suffix=suffix, delete=False) as tf:
        temp_image_path = pathlib.Path(tf.name)
    try:
        yield temp_image_path
    finally:
        try:
            tf.close()
            temp_image_path.unlink()
        except FileNotFoundError:
            pass
        except PermissionError:
            # Opps, windows issue
            pass


class TestInfo:
    test_name: str
    class_name: str

    def __init__(self, test_name: str, class_name: str):
        self.test_name = test_name
        self.class_name = class_name

    @staticmethod
    def get_test_info(
        request: FixtureRequest, suffix: str = "", prefix: str = ""
    ) -> "TestInfo":
        try:
            names = junitxml.mangle_testnames(  # type: ignore
                request.node.nodeid.split("::")
            )  #
        except AttributeError:
            # pytest>=2.9.0
            names = junitxml.mangle_test_address(request.node.nodeid)

        classname = ".".join(names[:-1])
        test_name = names[-1]
        return TestInfo(test_name, classname)


def get_test_info(
    request: FixtureRequest, suffix: str = "", prefix: str = ""
) -> TestInfo:
    return TestInfo.get_test_info(request, suffix, prefix)


def image_save(image: ImageFileType, path: PathType) -> None:
    if isinstance(image, Image):
        image.save(path)
    elif isinstance(image, (str, pathlib.Path)):
        if not os.path.exists(image):
            raise ValueError("Image maybe path. Path not exists!")
        shutil.copyfile(image, str(path))
    elif hasattr(image, "read"):
        with open(path, "wb") as f:
            image = typing.cast(typing.BinaryIO, image)
            shutil.copyfileobj(
                image, f
            )  # type: ignore # Workaround for python/mypy#8962
    elif isinstance(image, bytes):
        with open(path, "wb") as f:
            f.write(image)
    else:
        raise NotImplementedError()


def ensure_dirs(filepath: str) -> None:
    if not os.path.exists(filepath):
        file_dir = os.path.dirname(filepath)
        if not os.path.exists(file_dir):
            os.makedirs(file_dir)


# def pytest_assertrepr_compare(config, op, left, right):
#     print(op, left, right)
#     pass



from io import BytesIO
from typing import cast, Union, Optional

from PIL import Image
from diffimg import diff
from imgdiff import simple_highlight, tile_images

from pytest_image_diff._types import (
    PathOrFileType,
    ImageSize,
    OrientationType,
    ColorModeType,
    ImageType,
    ColorType,
)


class ImgDiffOpts:
    border: int = 0
    spacing: int = 3
    bgcolor: str = "#FFF"
    sepcolor: str = "#CCC"
    opacity: int = 64
    orientation: OrientationType = "auto"
    timeout: int = 10


def make_highlights(
    ref_path: PathOrFileType, image_path: PathOrFileType, diff_path: PathOrFileType
) -> ImageType:
    ref_img = Image.open(ref_path).convert("RGB")
    img = Image.open(image_path).convert("RGB")
    opts = ImgDiffOpts()
    mask1, mask2 = simple_highlight(ref_img, img, opts)
    img = tile_images(ref_img, img, mask1, mask2, opts)
    img.save(diff_path)
    return img


def resize_canvas(im: ImageType, new_size: ImageSize) -> BytesIO:
    new_image = Image.new(im.mode, new_size, (255, 0, 255))
    new_image.paste(im, (0, 0, *im.size))
    fp = BytesIO()
    new_image.save(fp, im.format)
    fp.seek(0)
    return fp


ALPHA_COLOR = (255, 0, 255)  # Magenta


def convert_color_mode(
    im: ImageType,
    color_mode: ColorModeType,
    alpha_color: Optional[ColorType] = ALPHA_COLOR,
) -> Union[ImageType, None]:
    if im.mode == color_mode:
        return None
    im = im.convert("RGBA")
    background = Image.new("RGBA", im.size, alpha_color or ALPHA_COLOR)

    alpha_image = Image.alpha_composite(background, im)
    alpha_image = alpha_image.convert(color_mode)

    return alpha_image


def _diff(
    ref_path: PathOrFileType,
    image_path: PathOrFileType,
    diff_path: PathOrFileType,
    color_mode: Union[ColorModeType, None] = None,
    alpha_color: ColorType = ALPHA_COLOR,
) -> float:
    ref_im = Image.open(ref_path)
    im = Image.open(image_path)

    if ref_im.size != im.size:
        new_w = ref_im.size[0] - im.size[0]
        new_h = ref_im.size[1] - im.size[1]
        ref_new_size = list(ref_im.size)
        im_new_size = list(im.size)
        if new_w < 0:
            ref_new_size[0] = im.size[0]
        elif new_w > 0:
            im_new_size[0] = ref_im.size[0]

        if new_h < 0:
            ref_new_size[1] = im.size[1]
        elif new_h > 0:
            im_new_size[1] = ref_im.size[1]

        _im_new_size: ImageSize = cast(ImageSize, tuple(im_new_size))
        _ref_new_size: ImageSize = cast(ImageSize, tuple(ref_new_size))

        if ref_im.size != ref_new_size:
            ref_path = resize_canvas(ref_im, _ref_new_size)
        if im.size != im_new_size:
            image_path = resize_canvas(im, _im_new_size)

    if color_mode:
        # Normalization color_mode
        _ref_im = convert_color_mode(
            ref_im, color_mode=color_mode, alpha_color=alpha_color
        )
        _im = convert_color_mode(im, color_mode=color_mode, alpha_color=alpha_color)
        if _ref_im:
            _ref_im.save(ref_path)
        if _im:
            _im.save(image_path)

    threshold: float = diff(
        ref_path, image_path, delete_diff_file=True, ignore_alpha=True
    )

    make_highlights(ref_path, image_path, diff_path)
    return threshold



import os
from typing import Optional, Generator, TYPE_CHECKING

import pytest
from typing_extensions import Protocol

from .helpers import temp_file

if TYPE_CHECKING:
    from .plugin import DiffCompareResult

try:
    # Check pytest-splinter
    from pytest_splinter.plugin import Browser
except ImportError:
    Browser = None

if Browser:
    from ._types import ImageRegressionCallableType, ImageDiffColorModeType, ColorType

    __all__ = ["screenshot_regression"]

    class ScreenshotRegressionCallableType(Protocol):
        def __call__(
            self,
            browser: Optional[Browser] = None,
            threshold: Optional[float] = None,
            suffix: Optional[str] = None,
            xpath: Optional[str] = "",
            throw_exception: Optional[bool] = None,
            color_mode: Optional[ImageDiffColorModeType] = None,
            alpha_color: Optional[ColorType] = None,
        ) -> "DiffCompareResult":
            pass

    @pytest.fixture(scope="function")
    def screenshot_regression(
        browser: Browser,
        image_regression: ImageRegressionCallableType,
        image_diff_threshold: float,
        image_diff_throw_exception: bool,
        image_diff_color_mode: ImageDiffColorModeType,
        image_diff_alpha_color: ColorType,
    ) -> Generator[ScreenshotRegressionCallableType, None, None]:
        """
        Check regression browser screenshot

        :param browser: optional, by default from `browser` fixture
        :param threshold: float, by default from `image_diff_threshold`
        :param suffix: str, need for multiple checks  by one test
        :param xpath: str, optional xpath expression to select an element to
                screenshot instead of page
        """
        default_browser = browser

        def _factory(
            browser: Optional[Browser] = None,
            threshold: Optional[float] = None,
            suffix: Optional[str] = "",
            xpath: Optional[str] = "",
            throw_exception: Optional[bool] = None,
            color_mode: Optional[ImageDiffColorModeType] = None,
            alpha_color: Optional[ColorType] = None,
        ) -> "DiffCompareResult":
            if browser is None:
                browser = default_browser

            if threshold is None:
                threshold = image_diff_threshold

            if throw_exception is None:
                throw_exception = image_diff_throw_exception

            if color_mode is None:
                color_mode = image_diff_color_mode
            if alpha_color is None:
                alpha_color = image_diff_alpha_color

            with temp_file(suffix=".png") as temp_image_path:
                screenshot_path = os.fspath(temp_image_path)

                if xpath:
                    # `unique_file=False` since we already have a temporary file
                    #
                    # Since an xpath screenshot composes its own file name,
                    # we need to give it the prefix and
                    # suffix as separate parameters. `:-4` for the path without extension,
                    # then suffix given manually.
                    browser.find_by_xpath(xpath).first.screenshot(
                        screenshot_path[:-4],
                        suffix=screenshot_path[-4:],
                        unique_file=False,
                        full=True,
                    )
                else:
                    browser.driver.save_screenshot(screenshot_path)

                result = image_regression(
                    screenshot_path,
                    threshold,
                    suffix,
                    throw_exception=throw_exception,
                    color_mode=color_mode,
                    alpha_color=alpha_color,
                )
                return result

        yield _factory



import pathlib
from typing import BinaryIO, Union, Tuple, Optional, TYPE_CHECKING, Literal
from typing_extensions import Protocol


from PIL.Image import Image as ImageType

if TYPE_CHECKING:
    from .plugin import DiffCompareResult

PathType = Union[str, pathlib.Path]
PathOrFileType = Union[PathType, bytes, BinaryIO]
ImageFileType = Union[ImageType, PathOrFileType]
ImageSize = Tuple[int, int]
ColorModeType = Literal["RGB", "RGBA"]
ImageDiffColorModeType = Union[ColorModeType, None]
ColorType = Tuple[int, int, int]


class ImageRegressionCallableType(Protocol):
    def __call__(
        self,
        image: ImageFileType,
        threshold: Optional[float] = None,
        suffix: Optional[str] = None,
        throw_exception: Optional[bool] = None,
        color_mode: Optional[ImageDiffColorModeType] = None,
        alpha_color: Optional[ColorType] = None,
    ) -> "DiffCompareResult":
        pass


class ImageDiffCallableType(Protocol):
    def __call__(
        self,
        image: ImageFileType,
        image_2: ImageFileType,
        threshold: Optional[float] = None,
        suffix: Optional[str] = None,
        throw_exception: Optional[bool] = None,
        color_mode: Optional[ImageDiffColorModeType] = None,
        alpha_color: Optional[ColorType] = None,
    ) -> "DiffCompareResult":
        pass


# ``opts.orientation`` can be 'lr' for left-and-right,
# 'tb' for top-and-bottom, or 'auto' for automatic.
OrientationType = Literal["auto", "lr", "tb"]
